/*
utility classes used to describe style from python

check the syntax.css file for classes related to the python syntax
(comma separators, keyword prefix before statements, colons before blocks, indentation,...)
*/


/* style unifier (everything is a flex div) */

div {
  display: flex;
  flex-direction: column;
}

/* minimal utility classes */
.row {
  flex-direction: row;
  /* force strings to be on a single line with no line returns
   * (without this setting, string literals do not match their value)
   *
   * can make the body larger than the screen size
   * the default div flex behavior also does not seem to wrap rows
   * */
  white-space: nowrap;
}

.gap {
  gap: 5px;
}

.block {
	margin-left: 30px;
}


/* used during development to notice specific elements */
.dbg-show {
  background-color: rgba(200,0,0, 0.5);
  min-width: 100px;
  min-height: 100px;
}
/* used for debugging inline text nodes */
.bg-red {
  background-color: rgba(200, 0, 0, 0.5);
}

